/*
    Copyright (c) 2002 Gregory Montoir

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <SDL.h>
#include "SDL_framerate.h"
#include "shared.h"
#include "saves.h"
#include "config.h"


#define SMS_SCREEN_WIDTH   256
#define SMS_SCREEN_HEIGHT  192
#define MACHINE_FPS        60
#define WINDOW_TITLE       "SMS+SDL (0.9.3R1)"



static int screen_width = SMS_SCREEN_WIDTH;
static int screen_height = SMS_SCREEN_HEIGHT;
static SDL_Surface *sdl_screen = NULL;
static SDL_Surface *sdl_bitmap = NULL;
static FPSmanager *sdl_fps = NULL;



void sdl_update_display_beginning()
{
  SDL_LockSurface(sdl_bitmap);
  bitmap.data = (unsigned char*)sdl_bitmap->pixels;
}


void sdl_update_display_ending()
{
  SDL_Rect rect_screen;
  int dx = 0;
  int dy = 0;
  /* center the bitmap in the screen if necessary */
  if(sdl_screen->w != sdl_bitmap->w || sdl_screen->h != sdl_bitmap->h) {
    dx = (sdl_screen->w - sdl_bitmap->w) >> 1; 
    dy = (sdl_screen->h - sdl_bitmap->h) >> 1;
    SDL_FillRect(sdl_screen, NULL, 0);
  }
  rect_screen.x = dx;
  rect_screen.y = dy;
  rect_screen.w = sdl_bitmap->w-dx;
  rect_screen.h = sdl_bitmap->h-dy;
  SDL_UnlockSurface(sdl_bitmap);
  SDL_BlitSurface(sdl_bitmap, NULL, sdl_screen, &rect_screen);
  SDL_UpdateRect(sdl_screen, 0, 0, 0, 0);
}



static void sdl_init_paddles()
{
  cfg.pad[0].up = SDLK_UP;
  cfg.pad[0].down = SDLK_DOWN;
  cfg.pad[0].left = SDLK_LEFT;
  cfg.pad[0].right = SDLK_RIGHT;
  cfg.pad[0].b1 = SDLK_c;
  cfg.pad[0].b2 = SDLK_v;
  cfg.pad[0].start = SDLK_RETURN;
}


static int sdl_update_input(SDLKey k, int p)
{
  switch(k) {
  case SDLK_F1:
    cfg.state_slot = ++cfg.state_slot % STATE_SLOTS_NB;
    break;
  case SDLK_F2:
    save_state(cfg.game_name, cfg.state_slot);
    break;
  case SDLK_F3:
    load_state(cfg.game_name, cfg.state_slot);
    break;
  case SDLK_ESCAPE:
    return 0;
  }
  if(k == cfg.pad[0].start) {
    if(p) input.system |= (IS_GG) ? INPUT_START : INPUT_PAUSE;
    else  input.system &= (IS_GG) ? INPUT_START : INPUT_PAUSE;
  }
  else if(k == cfg.pad[0].up) {
    if(p) input.pad[0] |= INPUT_UP;
    else  input.pad[0] &= ~INPUT_UP;
  }
  else if(k == cfg.pad[0].down) {
    if(p) input.pad[0] |= INPUT_DOWN;
    else  input.pad[0] &= ~INPUT_DOWN;
  }
  else if(k == cfg.pad[0].left) {
    if(p) input.pad[0] |= INPUT_LEFT;
    else  input.pad[0] &= ~INPUT_LEFT;
  }
  else if(k == cfg.pad[0].right) {
    if(p) input.pad[0] |= INPUT_RIGHT;
    else  input.pad[0] &= ~INPUT_RIGHT;
  }
  else if(k == cfg.pad[0].b1) {
    if(p) input.pad[0] |= INPUT_BUTTON1;
    else  input.pad[0] &= ~INPUT_BUTTON1;
  }
  else if(k == cfg.pad[0].b2) {
    if(p) input.pad[0] |= INPUT_BUTTON2;
    else  input.pad[0] &= ~INPUT_BUTTON2;
  }
  return 1;
}


int sdl_loop()
{
  SDL_Event event;
  if(SDL_PollEvent(&event)) {
    switch(event.type) {
    case SDL_QUIT:
      return 0;
    case SDL_KEYUP:
      return sdl_update_input(event.key.keysym.sym, 0);
    case SDL_KEYDOWN:
      return sdl_update_input(event.key.keysym.sym, 1);
    }
  }
  return 1;
}


static void sdl_calculate_speed()
{
  static Uint32 prev_ticks = 0;
  static int frames_rendered = 0;
  Uint32 cur_ticks = SDL_GetTicks();
  int diff = (cur_ticks - prev_ticks)/1000;
  if(diff < 1)
    ++frames_rendered;
  else {
    char title[100];
    int truefps = frames_rendered / diff;
    int percent = truefps * 100 / MACHINE_FPS;
    sprintf(title,"%s [%d%%]", WINDOW_TITLE, percent);
    SDL_WM_SetCaption(title, NULL);
    frames_rendered = 0;
    prev_ticks = SDL_GetTicks();
  }
}


void sdl_sync()
{
  sdl_calculate_speed();
  /* synchronize if user haven't disabled it */
  if(sdl_fps) SDL_framerateDelay(sdl_fps);
}



static struct {
  int pos;
  int len;
  Uint8* buf;
} audio;


static void audio_callback(void *userdata, Uint8 *stream, int len)
{
  memcpy(stream, audio.buf, len);
  memset(audio.buf, 0, audio.len);
  audio.pos = 0;
}


int sdl_init_sound(int samplerate)
{
  SDL_AudioSpec as;
  if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) {
    printf("ERROR: can't initialize SDL AUDIO SUBSYSTEM: %s.\n", SDL_GetError());
    return 0;
  }
  as.freq = samplerate;
  as.format = AUDIO_S16;
  as.channels = 2;
  as.samples = 4096;
  as.callback = audio_callback;
  as.userdata = 0;
  if(SDL_OpenAudio(&as, NULL) == -1) {
    printf("ERROR: can't open audio: %s.\n", SDL_GetError());
    return 0;
  }
  audio.buf = (Uint8*)malloc(as.size);
  if(!audio.buf) {
    printf("ERROR: can't allocate memory for sound.\n");
    return 0;
  }
  audio.len = as.size;
  audio.pos = 0;
  memset(audio.buf, 0, audio.len);
  cfg.soundrate = as.freq;
  
  printf("Initializing SDL Sound... %dHz 16bits %s.\n", as.freq, as.channels==2?"stereo":"mono");
  SDL_PauseAudio(0);
  return 1;
}


void sdl_update_sound_buffer()
{
  int s = snd.bufsize*sizeof(short);
  if(audio.pos+s*2 < audio.len) {
    SDL_LockAudio();
    memcpy(audio.buf+audio.pos, snd.buffer[0], s);
    audio.pos += s;
    memcpy(audio.buf+audio.pos, snd.buffer[1], s);
    audio.pos += s;
	SDL_UnlockAudio();
  }
}




int sdl_init_machine(int fullscreen, int throttle)
{
  Uint32 vidflags = SDL_SWSURFACE;

  if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER) < 0) {
    printf("ERROR: can't initialize SDL VIDEO|TIMER SUBSYSTEM: %s.\n", SDL_GetError());
    return 0;
  }

  if(fullscreen) {
    vidflags |= SDL_FULLSCREEN;
    screen_width = 320;
    screen_height = 200;
    SDL_ShowCursor(SDL_DISABLE);
  }
  sdl_screen = SDL_SetVideoMode(screen_width, screen_height, 0, vidflags);
  if(!sdl_screen) {
    printf("ERROR: can't set video mode (%dx%d): %s.\n", screen_width, screen_height, SDL_GetError());
    return 0;  	
  }
  
  /* Create the 16 bits bitmap surface (RGB 565) */
  sdl_bitmap = SDL_CreateRGBSurface(SDL_SWSURFACE, SMS_SCREEN_WIDTH, SMS_SCREEN_HEIGHT, 16, 0xF800, 0x07E0, 0x001F, 0);
  if(!sdl_bitmap) {
    printf("ERROR: can't create primary surface: %s.\n", SDL_GetError());
    return 0;  	  	
  }

  memset(&bitmap, 0, sizeof(t_bitmap));
  bitmap.width  = sdl_bitmap->w;
  bitmap.height = sdl_bitmap->h;
  bitmap.depth  = 16;
  bitmap.pitch  = sdl_bitmap->pitch;
  bitmap.data   = (unsigned char *)sdl_bitmap->pixels;

  SDL_WM_SetCaption(WINDOW_TITLE, NULL);

  sdl_init_paddles();
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  if(throttle) {
	sdl_fps = (FPSmanager*)malloc(sizeof(FPSmanager));
	if(sdl_fps) {
	  SDL_initFramerate(sdl_fps);
	  SDL_setFramerate(sdl_fps, MACHINE_FPS);
	}
  }
  return 1;
}


void sdl_trash_machine()
{
  if(cfg.soundrate != 0) { SDL_CloseAudio(); free(audio.buf); }
  if(sdl_screen) SDL_FreeSurface(sdl_screen);
  if(sdl_bitmap) SDL_FreeSurface(sdl_bitmap);
  if(sdl_fps) free(sdl_fps);
  SDL_Quit();
}










