
/*
    Copyright (c) 2002 Gregory Montoir

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "defines.h"
#include <SDL.h>
#include "shared.h"
#include "config.h"
#include "saves.h"
#include "SDL_framerate.h"
#ifdef SMSSDL_SDLSOUND
#include <SDL_thread.h>
#endif
#ifdef SMSSDL_SEALSOUND
#include "sealintf.h"
#endif
#ifdef SMSSDL_BILINEAR_FILTERING
#include "sdlbilinear.h"
#endif




/* VIDEO */


#define SMS_SCREEN_WIDTH   256
#define SMS_SCREEN_HEIGHT  192
#define GG_SCREEN_WIDTH    160
#define GG_SCREEN_HEIGHT   144



static SDL_Surface *sdl_screen = NULL;
static SDL_Surface *sdl_bitmap = NULL;
#ifdef SMSSDL_SCANLINES
static SDL_Surface *sdl_scanline = NULL;
#endif

static int used_bitmap_width;
static int used_bitmap_height;
static unsigned int frames_rendered = 0;


static int sdl_video_init()
{
  int screen_width, screen_height;
  Uint32 vidflags = SDL_SWSURFACE;

  /* make only resizeable window only if using bilinear filtering */

  screen_width  = used_bitmap_width  = (IS_GG) ? GG_SCREEN_WIDTH  : SMS_SCREEN_WIDTH;
  screen_height = used_bitmap_height = (IS_GG) ? GG_SCREEN_HEIGHT : SMS_SCREEN_HEIGHT;

  if(cfg.render_mode == FILTER_SCANLINES50) {
	/* SMS value even for game gear ... */
	screen_width = used_bitmap_width = SMS_SCREEN_WIDTH * 2;
	screen_height = used_bitmap_height = SMS_SCREEN_HEIGHT * 2;
  }
  else if(cfg.render_mode == FILTER_BILINEAR) 
    vidflags |= SDL_RESIZABLE;

  if(SDL_Init(SDL_INIT_VIDEO) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }

  if(cfg.start_fullscreen) {
    vidflags |= SDL_FULLSCREEN;
    screen_width = 640;
    screen_height = 480;
    SDL_ShowCursor(SDL_DISABLE);
  }
  sdl_screen = SDL_SetVideoMode(screen_width, screen_height, 0, vidflags);
  if(!sdl_screen) {
    printf("ERROR: can't set video mode (%dx%d): %s.\n", screen_width, screen_height, SDL_GetError());
    return 0;  	
  }
  
  /* Create the 16 bits bitmap surface (RGB 565) */
  /* Even if we are in GameGear mode, we need to allocate space for a buffer of 256x192 */
  sdl_bitmap = SDL_CreateRGBSurface(SDL_SWSURFACE, SMS_SCREEN_WIDTH, SMS_SCREEN_HEIGHT, 16, 0xF800, 0x07E0, 0x001F, 0);
  if(!sdl_bitmap) {
    printf("ERROR: can't create primary surface: %s.\n", SDL_GetError());
    return 0;  	  	
  }

  memset(&bitmap, 0, sizeof(t_bitmap));
  bitmap.width  = sdl_bitmap->w;
  bitmap.height = sdl_bitmap->h;
  bitmap.depth  = 16;
  bitmap.pitch  = sdl_bitmap->pitch;
  bitmap.data   = (unsigned char *)sdl_bitmap->pixels;

  SDL_WM_SetCaption(SMSSDL_TITLE, NULL);
  return 1;
}


static void sdl_video_resize(int w, int h)
{
  sdl_screen = SDL_SetVideoMode(w, h, 0, SDL_SWSURFACE|SDL_RESIZABLE);
}


static void sdl_video_close()
{
  if(sdl_screen) SDL_FreeSurface(sdl_screen);
  if(sdl_bitmap) SDL_FreeSurface(sdl_bitmap);
#ifdef SMSSDL_SCANLINES
  if(sdl_scanline) SDL_FreeSurface(sdl_scanline);
#endif
  printf("[INFO] Frames rendered = %lu.\n", frames_rendered);

}


static void sdl_video_take_screenshot()
{
  int status;
  char ssname[0x100];
  strcpy(ssname, cfg.game_name);
  sprintf(strrchr(ssname, '.'), "-%03d.bmp", cfg.current_screenshot);
  ++cfg.current_screenshot;
  SDL_LockSurface(sdl_screen);
  status = SDL_SaveBMP(sdl_screen, ssname);
  SDL_UnlockSurface(sdl_screen);
  if(status == 0)
	printf("[INFO] Screenshot written to '%s'.\n", ssname);
}



static void sdl_video_prepare_update()
{
  SDL_LockSurface(sdl_bitmap);
  bitmap.data = (unsigned char*)sdl_bitmap->pixels;
}

#ifdef SMSSDL_SCANLINES

static void sdl_update_scanlined_50_surface(const SDL_Surface* in, int dx, int dy)
{
  unsigned short *p, *q;
  int j;

  if(!sdl_scanline) {
    sdl_scanline = SDL_CreateRGBSurface(SDL_SWSURFACE, dx, dy, 16, 0xF800, 0x07E0, 0x001F, 0);
	SDL_FillRect(sdl_scanline, NULL, 0);
  }
	
  SDL_LockSurface(sdl_scanline);

  p = sdl_scanline->pixels;
  q = in->pixels;

#ifdef SMSSDL_MMX
  {
  	int _edx = in->h;
	int _ecx = in->pitch;
	int _ebx = sdl_scanline->w;

	__asm {
		emms

		mov edi,p
		mov esi,q
			
		mov edx,_edx
		xor eax,eax
	
		mov ecx,_ecx
		mov ebx,_ebx

		align 16
sum:
		movq mm0,qword ptr[esi+eax]
		movq mm1,mm0

		punpcklwd mm0,mm0
		add eax,8
		
		punpckhwd mm1,mm1
		movq qword ptr[edi+2*eax-16],mm0

		cmp eax,ebx
		movq qword ptr[edi+2*eax-8],mm1

		jnz sum
		xor eax,eax

		add esi,ecx
		dec edx

		lea edi,[edi+4*ecx]
		jnz sum

		emms
	}
  }

#else

  for(j = 0; j < in->h; ++j) {
    int i;
	for(i = 0; i < sdl_scanline->w; i += 2) {
      *(p+i)=*(p+i+1) = *(q+i/2);
	}

	p += sdl_scanline->pitch * 2 / sizeof(short);
	q += in->pitch / sizeof(short);
  }

#endif
  SDL_UnlockSurface(sdl_scanline);
}

#endif


static void sdl_video_blit_center(SDL_Surface* screen, SDL_Surface* buffer)
{
  /* center the bitmap in the screen only if necessary */
  if(screen->w != buffer->w || screen->h != buffer->h) {
    SDL_Rect rect_screen;
    int dx = (screen->w - buffer->w) >> 1;
    int dy = (screen->h - buffer->h) >> 1;
    SDL_FillRect(screen, NULL, 0);
    rect_screen.x = dx;
    rect_screen.y = dy;
    rect_screen.w = buffer->w - dx;
    rect_screen.h = buffer->h - dy;
 	SDL_BlitSurface(buffer, NULL, screen, &rect_screen);
  }
  else
	SDL_BlitSurface(buffer, NULL, screen, NULL);
  SDL_UpdateRect(sdl_screen, 0, 0, 0, 0);
}


static void sdl_video_finish_update()
{
  switch(cfg.render_mode) {
#ifdef SMSSDL_BILINEAR_FILTERING
  case FILTER_BILINEAR: {
    SDL_Surface *n;
    int beg_x = (IS_GG) ? 48 : 0; /* (SMS_W - GG_W) / 2 */
    int beg_y = (IS_GG) ? 24 : 1; /* (SMS_H - GG_H) / 2 */
    n = sdl_create_bilinear_fitered_surface_integer(sdl_bitmap, beg_x, beg_y, used_bitmap_width, used_bitmap_height, sdl_screen->w, sdl_screen->h);
    SDL_UnlockSurface(sdl_bitmap);
    if(n) {
	  SDL_BlitSurface(n, NULL, sdl_screen, NULL);
	  SDL_UpdateRect(sdl_screen, 0, 0, 0, 0);
	  SDL_FreeSurface(n);
	}
  }
  break;
#endif
#ifdef SMSSDL_SCANLINES
  case FILTER_SCANLINES50: {
	sdl_update_scanlined_50_surface(sdl_bitmap, used_bitmap_width, used_bitmap_height);
	SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
  break;
#endif
  case FILTER_NORMAL: {
	SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_bitmap);
  }
  break;
  } // switch(cfg.render_mode)
}


static void sdl_video_update()
{
  int skip_current_frame = 0;

  ++frames_rendered;

  if(cfg.frame_skip > 1)
    skip_current_frame = (frames_rendered % cfg.frame_skip == 0) ? 0 : 1;

  if(!skip_current_frame) {
    sdl_video_prepare_update();
    sms_frame(skip_current_frame);
    sdl_video_finish_update();
  }
  else
    sms_frame(0);
}



/* CONTROLS */

static void sdl_controls_init()
{
  cfg.pad[0].up = SDLK_UP;
  cfg.pad[0].down = SDLK_DOWN;
  cfg.pad[0].left = SDLK_LEFT;
  cfg.pad[0].right = SDLK_RIGHT;
  cfg.pad[0].b1 = SDLK_c;
  cfg.pad[0].b2 = SDLK_v;
  cfg.pad[0].start = SDLK_RETURN;

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}


static int sdl_controls_update_input(SDLKey k, int p)
{
  if(!p) {
	switch(k) {
	  case SDLK_F1:
		cfg.state_slot = ++cfg.state_slot % STATE_SLOTS_NB;
		printf("[INFO] Slot changed to #%d.\n", cfg.state_slot);
		break;
	  case SDLK_F2:
		if(save_state(cfg.game_name, cfg.state_slot))
  		  printf("[INFO] Saved state to slot #%d.\n", cfg.state_slot);
		break;
	  case SDLK_F3:
		if(load_state(cfg.game_name, cfg.state_slot))
		  printf("[INFO] Loaded state from slot #%d.\n", cfg.state_slot);
		break;
	  case SDLK_F4:
		sdl_video_take_screenshot();
		break;
	  case SDLK_ESCAPE:
		return 0;
	}
  }
  if(k == cfg.pad[0].start) {
    if(p) input.system |= (IS_GG) ? INPUT_START : INPUT_PAUSE;
    else  input.system &= (IS_GG) ? ~INPUT_START : ~INPUT_PAUSE;
  }
  else if(k == cfg.pad[0].up) {
    if(p) input.pad[0] |= INPUT_UP;
    else  input.pad[0] &= ~INPUT_UP;
  }
  else if(k == cfg.pad[0].down) {
    if(p) input.pad[0] |= INPUT_DOWN;
    else  input.pad[0] &= ~INPUT_DOWN;
  }
  else if(k == cfg.pad[0].left) {
    if(p) input.pad[0] |= INPUT_LEFT;
    else  input.pad[0] &= ~INPUT_LEFT;
  }
  else if(k == cfg.pad[0].right) {
    if(p) input.pad[0] |= INPUT_RIGHT;
    else  input.pad[0] &= ~INPUT_RIGHT;
  }
  else if(k == cfg.pad[0].b1) {
    if(p) input.pad[0] |= INPUT_BUTTON1;
    else  input.pad[0] &= ~INPUT_BUTTON1;
  }
  else if(k == cfg.pad[0].b2) {
    if(p) input.pad[0] |= INPUT_BUTTON2;
    else  input.pad[0] &= ~INPUT_BUTTON2;
  }
  return 1;
}


/* SYNC */

#define MACHINE_FPS   60


static FPSmanager *sdl_fps = NULL;
static Uint32 startingticks = 0;

int sdl_sync_init()
{
  if(SDL_InitSubSystem(SDL_INIT_TIMER) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }
  sdl_fps = (FPSmanager*)malloc(sizeof(FPSmanager));
  if(sdl_fps) {
    SDL_initFramerate(sdl_fps);
    SDL_setFramerate(sdl_fps, MACHINE_FPS);
  }
  startingticks = SDL_GetTicks();
  return 1;
}


void sdl_sync_update()
{
  /* synchronize if user hasn't disabled it */
  if(sdl_fps) SDL_framerateDelay(sdl_fps);
}


void sdl_sync_close()
{
  int playtime = SDL_GetTicks() - startingticks;
  int avgfps = frames_rendered * 1000 / playtime;
  printf("[INFO] Average FPS = %d (%d%%).\n", avgfps, avgfps * 100 / MACHINE_FPS); 
  printf("[INFO] Play time = %lu sec.\n", playtime / 1000);
  if(sdl_fps) free(sdl_fps);
}






/* SOUND */



#ifdef SMSSDL_SDLSOUND



typedef struct {
  Uint8* buf;
  unsigned int sizebuf;
  SDL_sem* sem_callback;
} sdlsound;

static sdlsound audio;



static void sdl_sound_callback(void *userdata, Uint8 *stream, int len)
{
  memcpy(stream, audio.buf, len);
  SDL_SemPost(audio.sem_callback);
}


static int sdl_sound_init(int freq)
{
  SDL_AudioSpec as_desired, as_obtained;
  if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) {
    printf("ERROR: can't initialize SDL AUDIO SUBSYSTEM: %s.\n", SDL_GetError());
	return 0;
  }
  as_desired.freq = freq;
  as_desired.format = AUDIO_S16;
  as_desired.channels = 2;
  as_desired.samples = snd.bufsize;
  as_desired.callback = sdl_sound_callback;
  as_desired.userdata = 0;
  if(SDL_OpenAudio(&as_desired, &as_obtained) == -1) {
    printf("ERROR: can't open audio: %s.\n", SDL_GetError());
    return 0;
  }
  if(as_obtained.size != snd.bufsize * 2 * sizeof(short)) {
	printf("ERROR: soundcard driver does not accept specified samples size.\n");
	return 0;
  }
  audio.sizebuf = as_desired.size;
  audio.buf = (Uint8*)malloc(audio.sizebuf);
  if(!audio.buf) {
    printf("ERROR: can't allocate enough memory for sound.\n");
    return 0;
  }
  memset(audio.buf, 0, audio.sizebuf);
  audio.sem_callback = SDL_CreateSemaphore(0);
  if(audio.sem_callback == NULL) {
	free(audio.buf);
    printf("ERROR: can't create semaphore for sound.\n");
    return 0;
  }
  SDL_PauseAudio(0);
  return 1;
}


static void sdl_sound_close()
{
  SDL_DestroySemaphore(audio.sem_callback);
  SDL_CloseAudio();
  free(audio.buf);
}


static void sdl_sound_update()
{
  Uint8* p;
  SDL_SemWait(audio.sem_callback);
  SDL_LockAudio();
  p = audio.buf;
  memcpy(p, snd.buffer[0], snd.bufsize * sizeof(short));
  p += snd.bufsize * sizeof(short);
  memcpy(p, snd.buffer[1], snd.bufsize * sizeof(short));
  SDL_UnlockAudio();
}



#endif // SMSSDL_SOUND




int sdl_main_init()
{
  printf("Initializing virtual console emulation... ");
  /* set up the virtual console emulation */
  system_init(cfg.sound_rate);
  load_sram(cfg.game_name);
  printf("Ok.\n");

  printf("Initializing SDL... ");
  if(SDL_Init(0) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }
  printf("Ok.\n");

  printf("Initializing SDL VIDEO SUBSYSTEM... ");
  if(!sdl_video_init())
	return 0;
  printf("Ok.\n");

  printf("Initializing SDL TIMER SUBSYSTEM... ");
  if(!sdl_sync_init())
	return 0;
  printf("Ok.\n");

  printf("Initializing SDL CONTROLS SUBSYSTEM... ");
  sdl_controls_init();
  printf("Ok.\n");

  if(!snd.enabled)
	cfg.sound_rate = 0;
  else if(cfg.sound_rate) {
	switch(cfg.audio_core) {
#ifdef SMSSDL_SDLSOUND
	case AUDIO_SDL:
	  printf("Initializing SDL SOUND SUBSYSTEM... ");
      if(!sdl_sound_init(cfg.sound_rate))
        return 0;
	  printf("Ok.\n");
	  break;
#endif
#ifdef SMSSDL_SEALSOUND
	case AUDIO_SEAL:
	  printf("Initializing SEAL SOUND SYSTEM... ");
	  if(!seal_sound_init(cfg.sound_rate))
		return 0;
	  printf("Ok.\n");
	  break;
#endif
	}
  }

#ifdef SMSSDL_BILINEAR_FILTERING
  if(cfg.render_mode == FILTER_BILINEAR)
	sdl_bilinear_filter_init();
#endif

  return 1;
}


void sdl_main_loop()
{
  printf("Starting emulation...\n");
  
  while(1) {

	/* pump SDL events */
    SDL_Event event;
    if(SDL_PollEvent(&event)) {
      switch(event.type) {
	  case SDL_VIDEORESIZE:
	    sdl_video_resize(event.resize.w, event.resize.h);
		break;
      case SDL_KEYUP:
        if(!sdl_controls_update_input(event.key.keysym.sym, 0))
		  return;
		break;
      case SDL_KEYDOWN:
        if(!sdl_controls_update_input(event.key.keysym.sym, 1))
		  return;
        break;
      case SDL_QUIT:
        return;
	  }
	}

	sdl_video_update();

    if(cfg.sound_rate) {
	  switch(cfg.audio_core) {
#ifdef SMSSDL_SDLSOUND
	  case AUDIO_SDL:
        sdl_sound_update();
	    break;
#endif
#ifdef SMSSDL_SEALSOUND
	  case AUDIO_SEAL:
	    osd_play_streamed_sample_16(0, snd.buffer[0], snd.bufsize * 2, cfg.sound_rate, 60, -100);
        osd_play_streamed_sample_16(1, snd.buffer[1], snd.bufsize * 2, cfg.sound_rate, 60,  100);
	    break;
#endif
	  }
	}

	if(cfg.throttle)
   	  sdl_sync_update();
  }
}




void sdl_main_close()
{
  printf("Stopping emulation...\n");
  
  if(cfg.sound_rate) {
	switch(cfg.audio_core) {
#ifdef SMSSDL_SDLSOUND
	case AUDIO_SDL:
      sdl_sound_close();
	  break;
#endif
#ifdef SMSSDL_SEALSOUND
	case AUDIO_SEAL:
	  seal_sound_close();
	  break;
#endif
	}
  }

  sdl_sync_close();
  sdl_video_close();
  SDL_Quit();
  /* shutdown the virtual console emulation */
  save_sram(cfg.game_name);
  system_shutdown();
}
