
/*
    Copyright (c) 2002 Gregory Montoir

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <SDL.h>



static Uint8 alphatable[256][256];


void sdl_bilinear_filter_init()
{
  /* to speed up the bilinear filtering routine */
  int i, j;
  for(j = 0; j < 256; ++j)
    for(i = 0; i < 256; ++i)
      alphatable[i][j] = (i * j) >> 8;
}


SDL_Surface* sdl_create_bilinear_fitered_surface_float(const SDL_Surface *in, int dx, int dy)
{
  SDL_Surface *n;
  /* Create the same surface as the first one (RGB565) */
  n = SDL_CreateRGBSurface(SDL_SWSURFACE, dx, dy, 16, 0xF800, 0x07E0, 0x001F, 0);
  if(n) {
	float fsx = (float)in->w / (float)dx;
	float fsy = (float)in->h / (float)dy;
	float v = 0.0f;
	int j;

	SDL_LockSurface(n);

	for(j = 0; j < dy - 1; ++j) {
	  int i;
	  float u = 0.0f;
	  int iv = (int)v;
    float decy = v - iv;

	  for(i = 0; i < dx; ++i) {
		int iu = (int)u;
		float decx = u - iu;
		float ul = (1.0f - decx) * (1.0f - decy);
		float ur = decx * (1.0f - decy);
		float ll = (1.0f - decx) * decy;
		float lr = decx * decy;

		Uint16 c;
		Uint8 r[5], g[5], b[5];

	    /* we assume the 'in' surface to be locked */
		Uint8* pIn  = (Uint8*)in->pixels + iu * in->format->BytesPerPixel + iv * in->pitch;
		Uint8* pOut = (Uint8*)n->pixels + i * n->format->BytesPerPixel + j * n->pitch;

		c = *(Uint16*)pIn;
		SDL_GetRGB(c, in->format, &r[0], &g[0], &b[0]);
		c = *(Uint16*)(pIn+2);
		SDL_GetRGB(c, in->format, &r[1], &g[1], &b[1]);
		c = *(Uint16*)(pIn+in->pitch);
		SDL_GetRGB(c, in->format, &r[2], &g[2], &b[2]);
		c = *(Uint16*)(pIn+in->pitch+2);
		SDL_GetRGB(c, in->format, &r[3], &g[3], &b[3]);
		r[4] = (Uint8)(ul * r[0] + ur * r[1] + ll * r[2] + lr * r[3]);
		g[4] = (Uint8)(ul * g[0] + ur * g[1] + ll * g[2] + lr * g[3]);
		b[4] = (Uint8)(ul * b[0] + ur * b[1] + ll * b[2] + lr * b[3]);
		*((Uint16*)pOut) = (Uint16)SDL_MapRGB(n->format, r[4], g[4], b[4]);
		u += fsx;
	  }
	  v += fsy;
	  
	}
	/* draw the last line */
	for(; j < dy; ++j) {
	  int i = 0;
	  for(i = 0; i < dx; ++i) {
		Uint8* pIn  = (Uint8*)in->pixels + (Uint8)(i * in->format->BytesPerPixel * fsx) + (Uint8)(j * in->pitch * fsy);
		Uint8* pOut = (Uint8*)n->pixels + i * n->format->BytesPerPixel + j * n->pitch;
		*((Uint16*)pOut) = *((Uint16*)pIn);
	  }
	}
	SDL_UnlockSurface(n);
  }
  return n;
}



SDL_Surface* sdl_create_bilinear_fitered_surface_integer(const SDL_Surface *in, int bx, int by, int ix, int iy, int dx, int dy)
{
  SDL_Surface *n;
  n = SDL_CreateRGBSurface(SDL_SWSURFACE, dx, dy, 16, 0xF800, 0x07E0, 0x001F, 0);
  if(n) {
    int scalex = (int)(65536*((float)ix / (float)dx));
    int scaley = (int)(65536*((float)iy / (float)dy));
    Uint32 sx, sy;
    int i, j;

    sx = sy = 0;
    SDL_LockSurface(n);
    for(j = 0; j < dy - 1; ++j) {
      for(i = 0; i < dx; ++i) {
        Uint8 u = (sx >> 8) & 0xff;
        Uint8 v = (sy >> 8) & 0xff;
        Uint8 ul = alphatable[255 - u][255 - v];
        Uint8 ur = alphatable[u][255 - v];
        Uint8 ll = alphatable[255 - u][v];
        Uint8 lr = alphatable[u][v];
        Uint16 c;
        Uint8 r[5], g[5], b[5];

        /* we assume the 'in' surface to be locked */
        Uint8* pIn  = (Uint8*)in->pixels + (bx + (sx >> 16)) * in->format->BytesPerPixel + (by + (sy >> 16)) * in->pitch;
        Uint8* pOut = (Uint8*)n->pixels + i * n->format->BytesPerPixel + j * n->pitch;

        c = *(Uint16*)pIn;
        SDL_GetRGB(c, in->format, &r[0], &g[0], &b[0]);
        c = *(Uint16*)(pIn+2);
        SDL_GetRGB(c, in->format, &r[1], &g[1], &b[1]);
        c = *(Uint16*)(pIn+in->pitch);
        SDL_GetRGB(c, in->format, &r[2], &g[2], &b[2]);
        c = *(Uint16*)(pIn+in->pitch+2);
        SDL_GetRGB(c, in->format, &r[3], &g[3], &b[3]);
        r[4] = ((ul * r[0]) >> 8) + ((ur * r[1]) >> 8) + ((ll * r[2]) >> 8) + ((lr * r[3]) >> 8);
        g[4] = ((ul * g[0]) >> 8) + ((ur * g[1]) >> 8) + ((ll * g[2]) >> 8) + ((lr * g[3]) >> 8);
        b[4] = ((ul * b[0]) >> 8) + ((ur * b[1]) >> 8) + ((ll * b[2]) >> 8) + ((lr * b[3]) >> 8);
        *((Uint16*)pOut) = (Uint16)SDL_MapRGB(n->format, r[4], g[4], b[4]);

        sx += scalex;
      }
      sx = 0;
      sy += scaley;
    }
    /* draw the last line */
    for(; j < dy; ++j) {
      int i = 0;
      for(i = 0; i < dx; ++i) {
        Uint8* pIn  = (Uint8*)in->pixels + (Uint8)(i * in->format->BytesPerPixel * (scalex >> 16)) + (Uint8)(j * in->pitch * (scaley >> 16));
        Uint8* pOut = (Uint8*)n->pixels + i * n->format->BytesPerPixel + j * n->pitch;
        *((Uint16*)pOut) = *((Uint16*)pIn);
      }
    }
    SDL_UnlockSurface(n);
  }
  return n;
}

