
/*
    Copyright (c) 2002 Gregory Montoir

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <SDL.h>
#include "defines.h"
#include "shared.h"
#include "config.h"
#include "saves.h"
#include "sdlbilinear.h"
#include "2xsai.h"
#if USE_WIN32
#include "audiowin32.h"
#endif



/* VIDEO */


#define SMS_SCREEN_WIDTH   256
#define SMS_SCREEN_HEIGHT  192
#define GG_SCREEN_WIDTH    160
#define GG_SCREEN_HEIGHT   144



static SDL_Surface *sdl_screen = NULL;
static SDL_Surface *sdl_bitmap = NULL;
static SDL_Surface *sdl_scanline = NULL;

static int used_bitmap_width;
static int used_bitmap_height;
static unsigned int frames_rendered = 0;


static int sdl_video_init()
{
  int screen_width, screen_height;
  Uint32 vidflags = SDL_SWSURFACE;

  screen_width  = used_bitmap_width  = (IS_GG) ? GG_SCREEN_WIDTH  : SMS_SCREEN_WIDTH;
  screen_height = used_bitmap_height = (IS_GG) ? GG_SCREEN_HEIGHT : SMS_SCREEN_HEIGHT;

  if(cfg.render_mode == FILTER_SCALE2XADVMAME || cfg.render_mode == FILTER_SCANLINES50 || cfg.render_mode == FILTER_SCANLINES50MMX || cfg.render_mode == FILTER_SCANLINES100) {
    /* SMS values even for game gear ... */
    screen_width = used_bitmap_width = SMS_SCREEN_WIDTH * 2;
    screen_height = used_bitmap_height = SMS_SCREEN_HEIGHT * 2;
    sdl_scanline = SDL_CreateRGBSurface(SDL_SWSURFACE, screen_width, screen_height, 16, 0xF800, 0x07E0, 0x001F, 0);
  }
  else if(cfg.render_mode == FILTER_2XSAI || cfg.render_mode == FILTER_SUPER2XSAI || cfg.render_mode == FILTER_SUPEREAGLE) {
    screen_width = used_bitmap_width = SMS_SCREEN_WIDTH * 2;
    screen_height = used_bitmap_height = SMS_SCREEN_HEIGHT * 2 - 12;
    sdl_scanline = SDL_CreateRGBSurface(SDL_SWSURFACE, screen_width, screen_height, 16, 0xF800, 0x07E0, 0x001F, 0);
  }
  else if(cfg.render_mode == FILTER_BILINEAR) 
    vidflags |= SDL_RESIZABLE;

  if(SDL_InitSubSystem(SDL_INIT_VIDEO) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }

  if(cfg.start_fullscreen) {
    vidflags |= SDL_FULLSCREEN;
    screen_width = 640;
    screen_height = 480;
    SDL_ShowCursor(SDL_DISABLE);
  }

  sdl_screen = SDL_SetVideoMode(screen_width, screen_height, 0, vidflags);
  if(!sdl_screen) {
    printf("ERROR: can't set video mode (%dx%d): %s.\n", screen_width, screen_height, SDL_GetError());
    return 0;
  }

  /* Create the 16 bits bitmap surface (RGB 565) */
  /* Even if we are in GameGear mode, we need to allocate space for a buffer of 256x192 */
  sdl_bitmap = SDL_CreateRGBSurface(SDL_SWSURFACE, SMS_SCREEN_WIDTH, SMS_SCREEN_HEIGHT, 16, 0xF800, 0x07E0, 0x001F, 0);
  if(!sdl_bitmap) {
    printf("ERROR: can't create primary surface: %s.\n", SDL_GetError());
    return 0;  	  	
  }

  memset(&bitmap, 0, sizeof(t_bitmap));
  bitmap.width  = sdl_bitmap->w;
  bitmap.height = sdl_bitmap->h;
  bitmap.depth  = 16;
  bitmap.pitch  = sdl_bitmap->pitch;
  bitmap.data   = (unsigned char *)sdl_bitmap->pixels;

  SDL_WM_SetCaption(SMSSDL_TITLE, NULL);

  if(cfg.render_mode == FILTER_SUPER2XSAI || cfg.render_mode == FILTER_SUPEREAGLE)
    Init_2xSaI(565);
  else if(cfg.render_mode == FILTER_BILINEAR)
    sdl_bilinear_filter_init();

  return 1;
}


static void sdl_video_resize(int w, int h)
{
  sdl_screen = SDL_SetVideoMode(w, h, 0, SDL_SWSURFACE|SDL_RESIZABLE);
}


static void sdl_video_close()
{
  if(sdl_screen) SDL_FreeSurface(sdl_screen);
  if(sdl_bitmap) SDL_FreeSurface(sdl_bitmap);
  if(cfg.render_mode == FILTER_SCANLINES50 || cfg.render_mode == FILTER_SCANLINES50MMX || cfg.render_mode == FILTER_SUPER2XSAI || cfg.render_mode == FILTER_SUPEREAGLE)
    if(sdl_scanline) SDL_FreeSurface(sdl_scanline);
  printf("[INFO] Frames rendered = %lu.\n", frames_rendered);
}


static void sdl_video_take_screenshot()
{
  int status;
  char ssname[0x100];

  strcpy(ssname, cfg.game_name);
  sprintf(strrchr(ssname, '.'), "-%03d.bmp", cfg.current_screenshot);
  ++cfg.current_screenshot;
  SDL_LockSurface(sdl_screen);
  status = SDL_SaveBMP(sdl_screen, ssname);
  SDL_UnlockSurface(sdl_screen);
  if(status == 0)
    printf("[INFO] Screenshot written to '%s'.\n", ssname);
}


static void sdl_video_prepare_update()
{
  SDL_LockSurface(sdl_bitmap);
  bitmap.data = (unsigned char*)sdl_bitmap->pixels;
}


/* AdvanceMAME Scale2x effect */
static void sdl_scale2x(const SDL_Surface* in)
{
  int i, j;
  short *p, *q;
  unsigned int nextlineIn, nextlineOut;

  SDL_LockSurface(sdl_scanline);

  nextlineIn = in->pitch / sizeof(short);
  p = in->pixels;

  nextlineOut = sdl_scanline->pitch / sizeof(short);
  q = sdl_scanline->pixels;

  /* paint first line */
  for(i = 0; i < in->w; ++i)
    *(q + (i << 1)) = *(q + (i << 1) + 1) = *(q + nextlineOut + (i << 1)) = *(q + nextlineOut + (i << 1) + 1) = *(p + i);

  q += nextlineOut << 1;
  p += nextlineIn;
  /* paint left and right */
  for(j = 1; j < in->h - 1; ++j) {
    *q = *(q + 1) = *(q + nextlineOut) = *(q + nextlineOut + 1) = *p;
    *(q + sdl_scanline->w - 2) = *(q + sdl_scanline->w - 1) = *(q + nextlineOut + sdl_scanline->w - 2) = *(q + nextlineOut + sdl_scanline->w - 1) = *(p + in->w);
    p += nextlineIn;
    q += nextlineOut << 1;
  }

  /* paint last line */
  for(i = 0; i < in->w; ++i)
    *(q + (i << 1)) = *(q + (i << 1) + 1) = *(q + nextlineOut + (i << 1)) = *(q + nextlineOut + (i << 1) + 1) = *(p + i);

  p = in->pixels;
  q = (short*)sdl_scanline->pixels + (nextlineOut << 1);
  for(j = 1; j < in->h - 1; ++j) {
    for(i = 1; i < in->w - 1; ++i) {
      short A = *(p + i - nextlineIn - 1);
      short B = *(p + i - nextlineIn);
      short C = *(p + i - nextlineIn + 1);
      short D = *(p + i - 1);
      short E = *(p + i );
      short F = *(p + i + 1);
      short G = *(p + i + nextlineIn - 1);
      short H = *(p + i + nextlineIn);
      short I = *(p + i + nextlineIn + 1);

      *(q + (i << 1)) = D == B && B != F && D != H ? D : E;
      *(q + (i << 1) + 1) = B == F && B != D && F != H ? F : E;
      *(q + (i << 1) + nextlineOut) = D == H && D != B && H != F ? D : E;
      *(q + (i << 1) + nextlineOut + 1) = H == F && D != H && B != F ? F : E;
    }
    p += nextlineIn;
    q += nextlineOut << 1;
  }
  SDL_UnlockSurface(sdl_scanline);
}


static void sdl_update_scanlined_100_surface(const SDL_Surface* in)
{
  unsigned short *p, *q;
  int j;
	
  SDL_LockSurface(sdl_scanline);

  p = sdl_scanline->pixels;
  q = in->pixels;

 for(j = 0; j < in->h; ++j) {
    int i;
	  for(i = 0; i < sdl_scanline->w; i += 2)
      *(p+i) = *(p+i+1) = *(q+i/2);
	  p += sdl_scanline->pitch * 2 / sizeof(short);
	  q += in->pitch / sizeof(short);
  }

  SDL_UnlockSurface(sdl_scanline);
}


static void sdl_update_scanlined_50_surface(const SDL_Surface* in, int dx, int dy)
{
  unsigned short *p, *q;
  int j;
	
  SDL_LockSurface(sdl_scanline);

  p = sdl_scanline->pixels;
  q = in->pixels;

  for(j = 0; j < sdl_scanline->h; ++j) {
    int i;
	for(i = 0; i < in->w; i++) {
	  if(j % 2 != 0 && j < sdl_scanline->h - 1) {
	    unsigned short p1 = *(q + i);
	    unsigned short p2 = *(q + i + in->pitch / sizeof(unsigned short));
	    // 0111 1011 1110 1111 == 0x7BEF
	    *(p + 2*i) = *(p + 2*i + 1) = ((p1 + p2) >> 2) & 0x7BEF; 
	  }
	  else 
	    *(p + 2*i) = *(p + 2*i + 1) = *(q + i);
	}
	p += sdl_scanline->pitch / sizeof(unsigned short);
	if(j % 2 != 0)
	  q += in->pitch / sizeof(unsigned short);
  }

  SDL_UnlockSurface(sdl_scanline);
}


#if USE_MMX

static void sdl_update_scanlinedMMX_50_surface(const SDL_Surface* in, int dx, int dy)
{
  extern void scanlinesMMX50(Uint8*,Uint8*,int,int,int);
  SDL_LockSurface(sdl_scanline);
#ifdef _WIN32
  scanlinesMMX50(sdl_scanline->pixels, in->pixels, in->h - 1, in->pitch, sdl_scanline->w / 8);
#else
  _scanlinesMMX50(sdl_scanline->pixels, in->pixels, in->h - 1, in->pitch, sdl_scanline->w / 8);
#endif
  SDL_UnlockSurface(sdl_scanline);
}

#endif


static void sdl_video_blit_center(SDL_Surface* screen, SDL_Surface* buffer)
{
  /* center the bitmap in the screen only if necessary */
  if(screen->w != buffer->w || screen->h != buffer->h) {
    SDL_Rect rect_screen;
    int dx = (screen->w - buffer->w) >> 1;
    int dy = (screen->h - buffer->h) >> 1;
    SDL_FillRect(screen, NULL, 0);
    rect_screen.x = dx;
    rect_screen.y = dy;
    rect_screen.w = buffer->w - dx;
    rect_screen.h = buffer->h - dy;
 	  SDL_BlitSurface(buffer, NULL, screen, &rect_screen);
  }
  else
	  SDL_BlitSurface(buffer, NULL, screen, NULL);
  SDL_UpdateRect(sdl_screen, 0, 0, 0, 0);
}


static void sdl_video_finish_update()
{
  if(cfg.render_mode == FILTER_BILINEAR) {
    SDL_Surface *n;
    int beg_x = (IS_GG) ? (SMS_SCREEN_WIDTH - GG_SCREEN_WIDTH) >> 1 : 0;
    int beg_y = (IS_GG) ? (SMS_SCREEN_HEIGHT - GG_SCREEN_HEIGHT) >> 1 : 0;
    n = sdl_create_bilinear_fitered_surface_integer(sdl_bitmap, beg_x, beg_y, used_bitmap_width, used_bitmap_height, sdl_screen->w, sdl_screen->h);
    SDL_UnlockSurface(sdl_bitmap);
    if(n) {
	    SDL_BlitSurface(n, NULL, sdl_screen, NULL);
	    SDL_UpdateRect(sdl_screen, 0, 0, 0, 0);
	    SDL_FreeSurface(n);
    }
	}
  else if(cfg.render_mode == FILTER_2XSAI || cfg.render_mode == FILTER_SUPER2XSAI || cfg.render_mode == FILTER_SUPEREAGLE) {
    Uint8* sx = (Uint8*)sdl_bitmap->pixels + sdl_bitmap->pitch * 3;
	  SDL_LockSurface(sdl_scanline);
    if(cfg.render_mode == FILTER_2XSAI)
      _2xSaI(sx, sdl_bitmap->pitch, sdl_scanline->pixels, sdl_scanline->pitch, sdl_bitmap->w, sdl_bitmap->h - 6);
    else if(cfg.render_mode == FILTER_SUPER2XSAI)
      Super2xSaI(sx, sdl_bitmap->pitch, sdl_scanline->pixels, sdl_scanline->pitch, sdl_bitmap->w, sdl_bitmap->h - 6);
    else if(cfg.render_mode == FILTER_SUPEREAGLE)
      SuperEagle(sx, sdl_bitmap->pitch, sdl_scanline->pixels, sdl_scanline->pitch, sdl_bitmap->w, sdl_bitmap->h - 6);
    SDL_UnlockSurface(sdl_scanline);
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
  else if(cfg.render_mode == FILTER_SCALE2XADVMAME) {
    sdl_scale2x(sdl_bitmap);
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
  else if(cfg.render_mode == FILTER_SCANLINES50) {
	  sdl_update_scanlined_50_surface(sdl_bitmap, used_bitmap_width, used_bitmap_height);
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
  else if(cfg.render_mode == FILTER_SCANLINES100) {
	  sdl_update_scanlined_100_surface(sdl_bitmap);
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
#if USE_MMX
  else if(cfg.render_mode == FILTER_SCANLINES50MMX) {
	  sdl_update_scanlinedMMX_50_surface(sdl_bitmap, used_bitmap_width, used_bitmap_height);
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_scanline);
  }
#endif
  else if(cfg.render_mode == FILTER_NORMAL) {
	  SDL_UnlockSurface(sdl_bitmap);
    sdl_video_blit_center(sdl_screen, sdl_bitmap);
  }
}


static void sdl_video_update()
{
  int skip_current_frame = 0;

  ++frames_rendered;

  if(cfg.frame_skip > 1)
    skip_current_frame = (frames_rendered % cfg.frame_skip == 0) ? 0 : 1;

  if(!skip_current_frame) {
    sdl_video_prepare_update();
    sms_frame(0);
    sdl_video_finish_update();
  }
  else
    sms_frame(1);
}



/* CONTROLS */

/* Joystick */
static SDL_Joystick *sdl_joy = NULL;
static int joy_commit_range = 3276;
static int Xstatus = 1, Ystatus = 1;

static void sdl_joystick_init()
{
  if(SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0)
   printf("ERROR: %s.\n", SDL_GetError());

  sdl_joy = SDL_JoystickOpen(cfg.joystick_number);
  if(sdl_joy) {
    /* Print some information */
    printf("Joystick (%s) has %d axes and %d buttons.\n", SDL_JoystickName(cfg.joystick_number), SDL_JoystickNumAxes(sdl_joy), SDL_JoystickNumButtons(sdl_joy));
    printf("Using button mapping I=%d II=%d START=%d.\n", cfg.joystick_button[0], cfg.joystick_button[1], cfg.joystick_button[2]);
    SDL_JoystickEventState(SDL_ENABLE);
  }
  else
    printf("ERROR: Could not open joystick: %s.\n", SDL_GetError());
}

static void sdl_joystick_update_input(SDL_Event event)
{
	int axisval;
  switch(event.type) {
		case SDL_JOYAXISMOTION:
		/*--------------------------------------------------*/
			switch (event.jaxis.axis) {
			case 0:	/* X axis */
				axisval = event.jaxis.value;
				if (axisval > joy_commit_range) {
					if (Xstatus == 2)
						break;
					if (Xstatus == 0) {
						input.pad[0] &= ~INPUT_LEFT;
					}
					input.pad[0] |= INPUT_RIGHT;
					Xstatus = 2;
					break;
				}
				if (axisval < -joy_commit_range) {
					if (Xstatus == 0)
						break;
					if (Xstatus == 2) {
						input.pad[0] &= ~INPUT_RIGHT;
					}
					input.pad[0] |= INPUT_LEFT;
					Xstatus = 0;
					break;
				}

				/* axis is centered */

				if (Xstatus == 2) {
					input.pad[0] &= ~INPUT_RIGHT;
				}
				if (Xstatus == 0) {
					input.pad[0] &= ~INPUT_LEFT;
				}
				Xstatus = 1;
				break;

			case 1:	/* Y axis */
				axisval = event.jaxis.value;
				if (axisval > joy_commit_range) {
					if (Ystatus == 2)
						break;
					if (Ystatus == 0) {
						input.pad[0] &= ~INPUT_UP;
					}
					input.pad[0] |= INPUT_DOWN;
					Ystatus = 2;
					break;
				}
				if (axisval < -joy_commit_range) {
					if (Ystatus == 0)
						break;
					if (Ystatus == 2) {
						input.pad[0] &= ~INPUT_DOWN;
					}
					input.pad[0] |= INPUT_UP;
					Ystatus = 0;
					break;
				}

				/* axis is centered */

				if (Ystatus == 2) {
					input.pad[0] &= ~INPUT_DOWN;
				}
				if (Ystatus == 0) {
					input.pad[0] &= ~INPUT_UP;
				}
				Ystatus = 1;
				break;

			}
			break;

		/*--------------------------------------------------*/
		case SDL_JOYBUTTONDOWN:
		/*--------------------------------------------------*/
			if (event.jbutton.button == cfg.joystick_button[0]) {
				input.pad[0] |= INPUT_BUTTON1;
			} else if (event.jbutton.button == cfg.joystick_button[1]) {
				input.pad[0] |= INPUT_BUTTON2;
			} else if (event.jbutton.button == cfg.joystick_button[2]) {
				input.system |= (IS_GG) ? INPUT_START : INPUT_PAUSE;
			}
			break;

		/*--------------------------------------------------*/
		case SDL_JOYBUTTONUP:
		/*--------------------------------------------------*/
			if (event.jbutton.button == cfg.joystick_button[0]) {
				input.pad[0] &= ~INPUT_BUTTON1;
			} else if (event.jbutton.button == cfg.joystick_button[1]) {
				input.pad[0] &= ~INPUT_BUTTON2;
			} else if (event.jbutton.button == cfg.joystick_button[2]) {
				input.system &= (IS_GG) ? ~INPUT_START : ~INPUT_PAUSE;
			}
			break;
	}
}


static void sdl_controls_init()
{
  cfg.pad[0].up = SDLK_UP;
  cfg.pad[0].down = SDLK_DOWN;
  cfg.pad[0].left = SDLK_LEFT;
  cfg.pad[0].right = SDLK_RIGHT;
  cfg.pad[0].b1 = SDLK_c;
  cfg.pad[0].b2 = SDLK_v;
  cfg.pad[0].start = SDLK_RETURN;

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}


static int sdl_controls_update_input(SDLKey k, int p)
{
  if(!p) {
	  switch(k) {
	  case SDLK_ESCAPE:
		  return 0;
	  case SDLK_F1:
		  sdl_video_take_screenshot();
		  break;
	  case SDLK_F2:
		  if(save_state(cfg.game_name, cfg.state_slot))
  		  printf("[INFO] Saved state to slot #%d.\n", cfg.state_slot);
		  break;
	  case SDLK_F3:
		  if(load_state(cfg.game_name, cfg.state_slot))
		    printf("[INFO] Loaded state from slot #%d.\n", cfg.state_slot);
		  break;
    case SDLK_F4:
    case SDLK_F5:
    case SDLK_F6:
    case SDLK_F7:
    case SDLK_F8:
      cfg.frame_skip = k - SDLK_F4 + 1;
      if(cfg.frame_skip == 1)
        printf("[INFO] Frame skip disabled.\n");
      else
        printf("[INFO] Frame skip set to %d.\n", cfg.frame_skip - 1);
      break;
	  case SDLK_0:
    case SDLK_1:
		case SDLK_2:
		case SDLK_3:
		case SDLK_4:
		case SDLK_5:
		case SDLK_6:
		case SDLK_7:
		case SDLK_8:
		case SDLK_9:
	    cfg.state_slot = k - SDLK_0;
		  printf("[INFO] Slot changed to #%d.\n", cfg.state_slot);
      break;
    }
  }
  if(k == cfg.pad[0].start) {
    if(p) input.system |= (IS_GG) ? INPUT_START : INPUT_PAUSE;
    else  input.system &= (IS_GG) ? ~INPUT_START : ~INPUT_PAUSE;
  }
  else if(k == cfg.pad[0].up) {
    if(p) input.pad[0] |= INPUT_UP;
    else  input.pad[0] &= ~INPUT_UP;
  }
  else if(k == cfg.pad[0].down) {
    if(p) input.pad[0] |= INPUT_DOWN;
    else  input.pad[0] &= ~INPUT_DOWN;
  }
  else if(k == cfg.pad[0].left) {
    if(p) input.pad[0] |= INPUT_LEFT;
    else  input.pad[0] &= ~INPUT_LEFT;
  }
  else if(k == cfg.pad[0].right) {
    if(p) input.pad[0] |= INPUT_RIGHT;
    else  input.pad[0] &= ~INPUT_RIGHT;
  }
  else if(k == cfg.pad[0].b1) {
    if(p) input.pad[0] |= INPUT_BUTTON1;
    else  input.pad[0] &= ~INPUT_BUTTON1;
  }
  else if(k == cfg.pad[0].b2) {
    if(p) input.pad[0] |= INPUT_BUTTON2;
    else  input.pad[0] &= ~INPUT_BUTTON2;
  }
  else if(k == SDLK_TAB) {
    if(p) input.system |= INPUT_HARD_RESET;
    else  input.system &= ~INPUT_HARD_RESET;
  }
  return 1;
}


/* SYNC */

#define MACHINE_FPS   60


static Uint32 startingticks = 0;
static Uint32 currentticks = 0;
static unsigned int frame_times[3] = { 16, 17, 17 };

int sdl_sync_init()
{
  if(SDL_InitSubSystem(SDL_INIT_TIMER) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }
  startingticks = SDL_GetTicks();
  return 1;
}


void sdl_sync_update()
{
  while(SDL_GetTicks() - currentticks < frame_times[frames_rendered % 3]);
  currentticks = SDL_GetTicks();
}


void sdl_sync_close()
{
  int playtime = (SDL_GetTicks() - startingticks) / 1000;
  int avgfps = frames_rendered / playtime;
  printf("[INFO] Average FPS = %d (%d%%).\n", avgfps, avgfps * 100 / MACHINE_FPS); 
  printf("[INFO] Play time = %lu sec.\n", playtime);
}






/* SOUND */

static char* audio_buffer;
static int audio_write_buffer, audio_read_buffer;
static int audio_buffer_size;

static void sdl_sound_callback(void *userdata, Uint8 *stream, int len)
{
  int start, end;

  if((audio_write_buffer - audio_read_buffer) < len) { 
	  audio_read_buffer += len;
	  return;
  }

  start = audio_read_buffer % audio_buffer_size;
  end = (start + len) % audio_buffer_size;

  if(end < start) {
    memcpy(stream, audio_buffer + start, audio_buffer_size - start);
    memcpy(stream + audio_buffer_size - start, audio_buffer, end);
  }
  else
    memcpy(stream, audio_buffer + start, len);

  audio_read_buffer += len;
}

static int sdl_sound_init(int freq)
{
  SDL_AudioSpec as_desired, as_obtained;

  if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) {
    printf("ERROR: can't initialize SDL AUDIO SUBSYSTEM: %s.\n", SDL_GetError());
	  return 0;
  }

  as_desired.freq = freq;
  as_desired.format = AUDIO_S16LSB;
  as_desired.channels = 2;
  as_desired.samples = 1;
  as_desired.callback = sdl_sound_callback;
  as_desired.userdata = 0;

  // "samples" has to be power of two
  while((as_desired.samples * 2) < snd.bufsize)
    as_desired.samples *= 2;

  if(SDL_OpenAudio(&as_desired, &as_obtained) == -1) {
    printf("ERROR: can't open audio: %s.\n", SDL_GetError());
    return 0;
  }

  if(as_desired.samples != as_obtained.samples) {
    printf("ERROR: soundcard driver does not accept specified samples size.\n");
    return 0;
  }

  audio_buffer_size = 6 * 2 * sizeof(short) * snd.bufsize;
  audio_buffer = (char*)malloc(audio_buffer_size);
  memset(audio_buffer, 0, audio_buffer_size);
  audio_write_buffer = audio_read_buffer = 0;
  return 1;
}

static void sdl_sound_close()
{
  SDL_PauseAudio(1);
  SDL_CloseAudio();
  free(audio_buffer);
}

static void sdl_sound_update()
{
  int start, end;
  int len = 2 * sizeof(short) * snd.bufsize;
  char *Buffer = (char*)malloc(len);

  memcpy(Buffer, (char*)snd.buffer[0], snd.bufsize * sizeof(short));
  Buffer += snd.bufsize * sizeof(short);
  memcpy(Buffer, (char*)snd.buffer[1], snd.bufsize * sizeof(short));
  Buffer -= snd.bufsize * sizeof(short);

  while((audio_write_buffer - audio_read_buffer) > (audio_buffer_size - len)) { SDL_Delay(1); }

  SDL_LockAudio();

  start = audio_write_buffer % audio_buffer_size;
  end = (start + len) % audio_buffer_size;

  if(end < start) {
    memcpy(audio_buffer + start, Buffer, audio_buffer_size - start);
    memcpy(audio_buffer, Buffer + audio_buffer_size - start, end);
  }
  else
    memcpy(audio_buffer + start, Buffer, len);

  audio_write_buffer += len;

  SDL_UnlockAudio();

  free(Buffer);
}





int sdl_main_init()
{
  printf("Initializing virtual console emulation... ");
  /* set up the virtual console emulation */
  system_init(cfg.sound_rate);
  if(cfg.use_sram)
    load_sram(cfg.game_name);
  printf("Ok.\n");

  printf("Initializing SDL... ");
  if(SDL_Init(0) < 0) {
    printf("ERROR: %s.\n", SDL_GetError());
    return 0;
  }
  printf("Ok.\n");

  printf("Initializing SDL VIDEO SUBSYSTEM... ");
  if(!sdl_video_init())
    return 0;
  printf("Ok.\n");

  printf("Initializing SDL TIMER SUBSYSTEM... ");
  if(!sdl_sync_init())
    return 0;
  printf("Ok.\n");

  printf("Initializing SDL CONTROLS SUBSYSTEM... ");
  sdl_controls_init();
  printf("Ok.\n");

  if(cfg.joystick_number >= 0) {
    printf("Initializing SDL JOYSTICK SUBSYSTEM... ");
    sdl_joystick_init();
    printf("Ok.\n");
  }

  if(!snd.enabled)
	  cfg.sound_rate = 0;
  else if(cfg.sound_rate) {
	  switch(cfg.audio_core) {
	  case AUDIO_SDL:
	    printf("Initializing SDL SOUND SUBSYSTEM... ");
      if(!sdl_sound_init(cfg.sound_rate))
        return 0;
	    printf("Ok.\n");
	    break;
#if USE_WIN32
	  case AUDIO_WIN32API:
	    printf("Initializing OS SOUND SYSTEM... ");
	    if(!win32api_sound_init(cfg.sound_rate))
		    return 0;
	    printf("Ok.\n");
	    break;
#endif
    }
  }
  return 1;
}


void sdl_main_loop()
{
  printf("Starting emulation...\n");

  if(cfg.sound_rate && cfg.audio_core == AUDIO_SDL)
    SDL_PauseAudio(0);
  
  while(1) {
    /* pump SDL events */
    SDL_Event event;
    if(SDL_PollEvent(&event)) {
      switch(event.type) {
      case SDL_VIDEORESIZE:
        sdl_video_resize(event.resize.w, event.resize.h);
        break;
      case SDL_KEYUP:
        if(!sdl_controls_update_input(event.key.keysym.sym, 0))
          return;
        break;
      case SDL_KEYDOWN:
        if(!sdl_controls_update_input(event.key.keysym.sym, 1))
          return;
        break;
      case SDL_JOYAXISMOTION:
      case SDL_JOYBUTTONDOWN:
      case SDL_JOYBUTTONUP:
        sdl_joystick_update_input(event);
        break;
      case SDL_QUIT:
        return;
      }
    }

    sdl_video_update();

    if(cfg.sound_rate) {
      switch(cfg.audio_core) {
      case AUDIO_SDL:
        sdl_sound_update();
        break;
#if USE_WIN32
      case AUDIO_WIN32API:
        win32api_sound_update();
        break;
#endif
      }
    }

    if(cfg.throttle && (!cfg.sound_rate || cfg.audio_core != AUDIO_SDL))
      sdl_sync_update();
  }
}



void sdl_main_close()
{
  printf("Stopping emulation...\n");
  
  if(cfg.sound_rate) {
	  switch(cfg.audio_core) {
	  case AUDIO_SDL:
      sdl_sound_close();
	    break;
#if USE_WIN32
	  case AUDIO_WIN32API:
	    win32api_sound_close();
	    break;
#endif
	  }
  }

  sdl_video_close();
  sdl_sync_close();
  SDL_Quit();
  /* shutdown the virtual console emulation */
  save_sram(cfg.game_name);
  system_shutdown();
}
